
// iR3000a.c assembly routines

.intel_syntax noprefix

//////////////////////////////////////////////////////////////////////////
// Note that iR3000A.S and iR5900.S asm code is now identical.  Only some
// function names and the following two defines should ever differ:

#define REGINFO psxRegs
#define RECLUT psxRecLUT
#define PCOFFSET 0x208 // this must always match what Pcsx2 displays at startup

//////////////////////////////////////////////////////////////////////////
// Preprocessor Mess!

.extern REGINFO
.extern RECLUT
.extern iopRecRecompile

#define BLOCKTYPE_STARTPC	4		// startpc offset
#define BLOCKTYPE_DELAYSLOT	1		// if bit set, delay slot

#define BASEBLOCK_SIZE 2 // in dwords

//////////////////////////////////////////////////////////////////////////
// The address for all cleared blocks.  It recompiles the current pc and then
// dispatches to the recompiled block address.
.global iopJITCompile
iopJITCompile:

	mov esi, dword ptr [REGINFO + PCOFFSET]
	push esi
	call iopRecRecompile
	add esp, 4
	mov ebx, esi
	shr esi, 16
	mov ecx, dword ptr [RECLUT+esi*4]
	jmp dword ptr [ecx+ebx]

.global iopJITCompileInBlock
iopJITCompileInBlock:
	
	jmp iopJITCompile
	
//////////////////////////////////////////////////////////////////////////
// Recompiles the next block, and links the old block directly to it.
// This is a on-shot execution for any block which uses it.  Once the block
// has been statically linked to the new block, this function will be bypassed
//
// edx - jump address to modify
.globl iopDispatcher
iopDispatcher:

	# calc PC_GETBLOCK
	# ((BASEBLOCK*)(reclut[((u32)(x)) >> 16] + (x)*(sizeof(BASEBLOCK)/4)))

	mov eax, dword ptr [REGINFO + PCOFFSET]
	mov ebx, eax
	shr eax, 16
	mov ecx, dword ptr [RECLUT+eax*4]
	mov eax, dword ptr [ecx+ebx]

	cmp eax, offset iopJITCompile
	je Dispatch_notcompiled
	cmp eax, offset iopJITCompileInBlock
	je Dispatch_notcompiled
	lea ebx, [eax-4]
	sub ebx, edx
	mov dword ptr [edx], ebx
	jmp eax

	.align 16
Dispatch_notcompiled:
	mov esi, edx
	lea edi, [ecx+ebx]
	push ebx
	call iopRecRecompile
	add esp, 4

	mov eax, dword ptr [edi]
	lea ebx, [eax-4]
	sub ebx, esi
	mov dword ptr [esi], ebx
	jmp eax


//////////////////////////////////////////////////////////////////////////
// edx -  baseblock->startpc
// stack - x86Ptr
.globl iopDispatcherClear
iopDispatcherClear:
	mov [REGINFO + PCOFFSET], edx

	# calc PC_GETBLOCK
	# ((BASEBLOCK*)(reclut[((u32)(x)) >> 16] + (x)*(sizeof(BASEBLOCK)/4)))

	mov ebx, edx
	shr edx, 16
	mov ecx, dword ptr [RECLUT+edx*4]
	mov eax, dword ptr [ecx+ebx]

	cmp eax, offset iopJITCompile
	je Clear_notcompiled
	cmp eax, offset iopJITCompileInBlock
	je Clear_notcompiled
	add esp, 4
	jmp eax

	.align 16
Clear_notcompiled:
	lea edi, [ecx+ebx]
	push ebx
	call iopRecRecompile
	add esp, 4
	mov eax, dword ptr [edi]

	pop ecx
	mov byte ptr [ecx], 0xe9 // jmp32
	lea ebx, [eax-5]
	sub ebx, ecx
	mov dword ptr [ecx+1], ebx

	jmp eax


//////////////////////////////////////////////////////////////////////////
// called when jumping to variable pc address.

.globl iopDispatcherReg
iopDispatcherReg:

	mov eax, dword ptr [REGINFO + PCOFFSET]
	mov ebx, eax
	shr eax, 16
	mov ecx, dword ptr [RECLUT+eax*4]
	jmp dword ptr [ecx+ebx]
