/*  Pcsx2 - Pc Ps2 Emulator
*  Copyright (C) 2009  Pcsx2-Playground Team
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*  
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
*/

#pragma once
#ifdef PCSX2_MICROVU

//------------------------------------------------------------------
// Micro VU Micromode Lower instructions
//------------------------------------------------------------------

microVUf(void) mVU_DIV() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		//u8 *pjmp;, *pjmp1;
		u32 *ajmp32, *bjmp32;

		getReg5(xmmFs, _Fs_, _Fsf_);
		getReg5(xmmFt, _Ft_, _Ftf_);

		//AND32ItoM(VU_VI_ADDR(REG_STATUS_FLAG, 2), 0xFCF); // Clear D/I flags

		// FT can be zero here! so we need to check if its zero and set the correct flag.
		SSE_XORPS_XMM_to_XMM(xmmT1, xmmT1); // Clear xmmT1
		SSE_CMPEQPS_XMM_to_XMM(xmmT1, xmmFt); // Set all F's if each vector is zero
		SSE_MOVMSKPS_XMM_to_R32(gprT1, xmmT1); // Move the sign bits of the previous calculation

		AND32ItoR(gprT1, 1);  // Grab "Is Zero" bits from the previous calculation
		ajmp32 = JZ32(0); // Skip if none are
			//SSE_XORPS_XMM_to_XMM(xmmT1, xmmT1); // Clear xmmT1
			//SSE_CMPEQPS_XMM_to_XMM(xmmT1, xmmFs); // Set all F's if each vector is zero
			//SSE_MOVMSKPS_XMM_to_R32(gprT1, xmmT1); // Move the sign bits of the previous calculation

			//AND32ItoR(gprT1, 1);  // Grab "Is Zero" bits from the previous calculation
			//pjmp = JZ8(0);
			//	OR32ItoM( VU_VI_ADDR(REG_STATUS_FLAG, 2), 0x410 ); // Set invalid flag (0/0)
			//	pjmp1 = JMP8(0);
			//x86SetJ8(pjmp);
			//	OR32ItoM( VU_VI_ADDR(REG_STATUS_FLAG, 2), 0x820 ); // Zero divide (only when not 0/0)
			//x86SetJ8(pjmp1);

			SSE_XORPS_XMM_to_XMM(xmmFs, xmmFt);
			SSE_ANDPS_M128_to_XMM(xmmFs, (uptr)mVU_signbit);
			SSE_ORPS_M128_to_XMM(xmmFs, (uptr)mVU_maxvals); // If division by zero, then xmmFs = +/- fmax

			bjmp32 = JMP32(0);
		x86SetJ32(ajmp32);

		SSE_DIVSS_XMM_to_XMM(xmmFs, xmmFt);
		mVUclamp1<vuIndex>(xmmFs, xmmFt, 8);

		x86SetJ32(bjmp32);

		mVUunpack_xyzw<vuIndex>(xmmFs, xmmFs, 0);
		mVUmergeRegs<vuIndex>(xmmPQ, xmmFs, writeQ ? 4 : 8);
	}
}
microVUf(void) mVU_SQRT() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		//u8* pjmp;
		getReg5(xmmFt, _Ft_, _Ftf_);

		//AND32ItoM(VU_VI_ADDR(REG_STATUS_FLAG, 2), 0xFCF); // Clear D/I flags
		/* Check for negative sqrt */
		//SSE_MOVMSKPS_XMM_to_R32(gprT1, xmmFt);
		//AND32ItoR(gprT1, 1);  //Check sign
		//pjmp = JZ8(0); //Skip if none are
		//	OR32ItoM(VU_VI_ADDR(REG_STATUS_FLAG, 2), 0x410); // Invalid Flag - Negative number sqrt
		//x86SetJ8(pjmp);

		SSE_ANDPS_M128_to_XMM(xmmFt, (uptr)mVU_absclip); // Do a cardinal sqrt
		if (CHECK_VU_OVERFLOW) SSE_MINSS_M32_to_XMM(xmmFt, (uptr)mVU_maxvals); // Clamp infinities (only need to do positive clamp since xmmFt is positive)
		SSE_SQRTSS_XMM_to_XMM(xmmFt, xmmFt);
		mVUunpack_xyzw<vuIndex>(xmmFt, xmmFt, 0);
		mVUmergeRegs<vuIndex>(xmmPQ, xmmFt, writeQ ? 4 : 8);
	}
}
microVUf(void) mVU_RSQRT() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		u8 *ajmp8, *bjmp8;

		getReg5(xmmFs, _Fs_, _Fsf_);
		getReg5(xmmFt, _Ft_, _Ftf_);

		//AND32ItoM(VU_VI_ADDR(REG_STATUS_FLAG, 2), 0xFCF); // Clear D/I flags
		/* Check for negative divide */
		//SSE_MOVMSKPS_XMM_to_R32(gprT1, xmmT1);
		//AND32ItoR(gprT1, 1);  //Check sign
		//ajmp8 = JZ8(0); //Skip if none are
		//	OR32ItoM(VU_VI_ADDR(REG_STATUS_FLAG, 2), 0x410); // Invalid Flag - Negative number sqrt
		//x86SetJ8(ajmp8);

		SSE_ANDPS_M128_to_XMM(xmmFt, (uptr)mVU_absclip); // Do a cardinal sqrt
		SSE_SQRTSS_XMM_to_XMM(xmmFt, xmmFt);

		// Ft can still be zero here! so we need to check if its zero and set the correct flag.
		SSE_XORPS_XMM_to_XMM(xmmT1, xmmT1); // Clear t1reg
		SSE_CMPEQSS_XMM_to_XMM(xmmT1, xmmFt); // Set all F's if each vector is zero

		SSE_MOVMSKPS_XMM_to_R32(gprT1, xmmT1); // Move the sign bits of the previous calculation

		AND32ItoR(gprT1, 1);  // Grab "Is Zero" bits from the previous calculation
		ajmp8 = JZ8(0); // Skip if none are
			//OR32ItoM(VU_VI_ADDR(REG_STATUS_FLAG, 2), 0x820); // Zero divide flag
			SSE_ANDPS_M128_to_XMM(xmmFs, (uptr)mVU_signbit);
			SSE_ORPS_M128_to_XMM(xmmFs, (uptr)mVU_maxvals); // EEREC_TEMP = +/-Max
			bjmp8 = JMP8(0);
		x86SetJ8(ajmp8);
			SSE_DIVSS_XMM_to_XMM(xmmFs, xmmFt);
			mVUclamp1<vuIndex>(xmmFs, xmmFt, 8);
		x86SetJ8(bjmp8);

		mVUunpack_xyzw<vuIndex>(xmmFs, xmmFs, 0);
		mVUmergeRegs<vuIndex>(xmmPQ, xmmFs, writeQ ? 4 : 8);
	}
}

#define EATANhelper(addr) {						\
	SSE_MULSS_XMM_to_XMM(xmmT1, xmmFs);			\
	SSE_MULSS_XMM_to_XMM(xmmT1, xmmFs);			\
	SSE_MOVAPS_XMM_to_XMM(xmmFt, xmmT1);		\
	SSE_MULSS_M32_to_XMM(xmmFt, (uptr)addr);	\
	SSE_ADDSS_XMM_to_XMM(xmmPQ, xmmFt);			\
}
microVUt(void) mVU_EATAN_() {
	microVU* mVU = mVUx;

	// ToDo: Can Be Optimized Further? (takes approximately (~115 cycles + mem access time) on a c2d)
	SSE_MOVSS_XMM_to_XMM(xmmPQ, xmmFs);
	SSE_MULSS_M32_to_XMM(xmmPQ, (uptr)mVU_T1);
	SSE_MOVAPS_XMM_to_XMM(xmmT1, xmmFs);

	EATANhelper(mVU_T2);
	EATANhelper(mVU_T3);
	EATANhelper(mVU_T4);
	EATANhelper(mVU_T5);
	EATANhelper(mVU_T6);
	EATANhelper(mVU_T7);
	EATANhelper(mVU_T8);

	SSE_ADDSS_M32_to_XMM(xmmPQ, (uptr)mVU_Pi4);
	SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6);
}
microVUf(void) mVU_EATAN() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		getReg5(xmmFs, _Fs_, _Fsf_);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip xmmPQ to get Valid P instance

		SSE_MOVSS_XMM_to_XMM(xmmPQ, xmmFs);
		SSE_SUBSS_M32_to_XMM(xmmFs, (uptr)mVU_one);
		SSE_ADDSS_M32_to_XMM(xmmPQ, (uptr)mVU_one);
		SSE_DIVSS_XMM_to_XMM(xmmFs, xmmPQ);

		mVU_EATAN_<vuIndex>();
	}
}
microVUf(void) mVU_EATANxy() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		getReg6(xmmFt, _Fs_);
		SSE2_PSHUFD_XMM_to_XMM(xmmFs, xmmFt, 0x01);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip xmmPQ to get Valid P instance

		SSE_MOVSS_XMM_to_XMM(xmmPQ, xmmFs);
		SSE_SUBSS_M32_to_XMM(xmmFs, (uptr)mVU_one);
		SSE_ADDSS_XMM_to_XMM(xmmFt, xmmPQ);
		SSE_DIVSS_XMM_to_XMM(xmmFs, xmmFt);

		mVU_EATAN_<vuIndex>();
	}
}
microVUf(void) mVU_EATANxz() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		getReg6(xmmFt, _Fs_);
		SSE2_PSHUFD_XMM_to_XMM(xmmFs, xmmFt, 0x02);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip xmmPQ to get Valid P instance

		SSE_MOVSS_XMM_to_XMM(xmmPQ, xmmFs);
		SSE_SUBSS_XMM_to_XMM(xmmFs, xmmFt);
		SSE_ADDSS_XMM_to_XMM(xmmFt, xmmPQ);
		SSE_DIVSS_XMM_to_XMM(xmmFs, xmmFt);

		mVU_EATAN_<vuIndex>();
	}
}
#define eexpHelper(addr) {  \
	SSE_MULSS_XMM_to_XMM(xmmT1, xmmFs);  \
	SSE_MOVAPS_XMM_to_XMM(xmmFt, xmmT1);  \
	SSE_MULSS_M32_to_XMM(xmmFt, (uptr)addr);  \
	SSE_ADDSS_XMM_to_XMM(xmmPQ, xmmFt);  \
}
microVUf(void) mVU_EEXP() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		getReg5(xmmFs, _Fs_, _Fsf_);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip xmmPQ to get Valid P instance
		SSE_MOVSS_XMM_to_XMM(xmmPQ, xmmFs);
		SSE_MULSS_M32_to_XMM(xmmPQ, (uptr)mVU_E1);
		SSE_ADDSS_M32_to_XMM(xmmPQ, (uptr)mVU_one);
		
		SSE_MOVAPS_XMM_to_XMM(xmmFt, xmmFs);
		SSE_MULSS_XMM_to_XMM(xmmFt, xmmFs);
		SSE_MOVAPS_XMM_to_XMM(xmmT1, xmmFt);
		SSE_MULSS_M32_to_XMM(xmmFt, (uptr)mVU_E2);
		SSE_ADDSS_XMM_to_XMM(xmmPQ, xmmFt);

		eexpHelper(mVU_E3);
		eexpHelper(mVU_E4);
		eexpHelper(mVU_E5);
		
		SSE_MULSS_XMM_to_XMM(xmmT1, xmmFs);
		SSE_MULSS_M32_to_XMM(xmmT1, (uptr)mVU_E6);
		SSE_ADDSS_XMM_to_XMM(xmmPQ, xmmT1);
		SSE_MULSS_XMM_to_XMM(xmmPQ, xmmPQ);
		SSE_MULSS_XMM_to_XMM(xmmPQ, xmmPQ);
		SSE_MOVSS_M32_to_XMM(xmmT1, (uptr)mVU_one);
		SSE_DIVSS_XMM_to_XMM(xmmT1, xmmPQ);
		SSE_MOVSS_XMM_to_XMM(xmmPQ, xmmT1);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip back
	}
}
microVUt(void) mVU_sumXYZ() { 
	// regd.x =  x ^ 2 + y ^ 2 + z ^ 2
	if( cpucaps.hasStreamingSIMD4Extensions ) {
		SSE4_DPPS_XMM_to_XMM(xmmFs, xmmFs, 0x71);
		SSE_MOVSS_XMM_to_XMM(xmmPQ, xmmFs);
	}
	else {
		SSE_MULPS_XMM_to_XMM(xmmFs, xmmFs); // wzyx ^ 2
		SSE_MOVSS_XMM_to_XMM(xmmPQ, xmmFs);
		SSE2_PSHUFD_XMM_to_XMM(xmmFs, xmmFs, 0xe1); // wzyx -> wzxy
		SSE_ADDSS_XMM_to_XMM(xmmPQ, xmmFs); // x ^ 2 + y ^ 2
		SSE2_PSHUFD_XMM_to_XMM(xmmFs, xmmFs, 0xD2); // wzxy -> wxyz
		SSE_ADDSS_XMM_to_XMM(xmmPQ, xmmFs); // x ^ 2 + y ^ 2 + z ^ 2
	}
}
microVUf(void) mVU_ELENG() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		getReg6(xmmFs, _Fs_);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip xmmPQ to get Valid P instance
		mVU_sumXYZ<vuIndex>();
		SSE_SQRTSS_XMM_to_XMM(xmmPQ, xmmPQ);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip back
	}
}
microVUf(void) mVU_ERCPR() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		getReg5(xmmFs, _Fs_, _Fsf_);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip xmmPQ to get Valid P instance
		SSE_MOVSS_XMM_to_XMM(xmmPQ, xmmFs);
		SSE_MOVSS_M32_to_XMM(xmmFs, (uptr)mVU_one);
		SSE_DIVSS_XMM_to_XMM(xmmFs, xmmPQ);
		SSE_MOVSS_XMM_to_XMM(xmmPQ, xmmFs);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip back
	}
}
microVUf(void) mVU_ERLENG() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		getReg6(xmmFs, _Fs_);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip xmmPQ to get Valid P instance
		mVU_sumXYZ<vuIndex>();
		SSE_SQRTSS_XMM_to_XMM(xmmPQ, xmmPQ);
		SSE_MOVSS_M32_to_XMM(xmmFs, (uptr)mVU_one);
		SSE_DIVSS_XMM_to_XMM(xmmFs, xmmPQ);
		SSE_MOVSS_XMM_to_XMM(xmmPQ, xmmFs);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip back
	}
}
microVUf(void) mVU_ERSADD() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		getReg6(xmmFs, _Fs_);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip xmmPQ to get Valid P instance
		mVU_sumXYZ<vuIndex>();
		//SSE_RCPSS_XMM_to_XMM(xmmPQ, xmmPQ); // Lower Precision is bad?
		SSE_MOVSS_M32_to_XMM(xmmFs, (uptr)mVU_one);
		SSE_DIVSS_XMM_to_XMM(xmmFs, xmmPQ);
		SSE_MOVSS_XMM_to_XMM(xmmPQ, xmmFs);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip back
	}
}
microVUf(void) mVU_ERSQRT() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		getReg5(xmmFs, _Fs_, _Fsf_);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip xmmPQ to get Valid P instance
		SSE_SQRTSS_XMM_to_XMM(xmmPQ, xmmFs);
		SSE_MOVSS_M32_to_XMM(xmmFs, (uptr)mVU_one);
		SSE_DIVSS_XMM_to_XMM(xmmFs, xmmPQ);
		SSE_MOVSS_XMM_to_XMM(xmmPQ, xmmFs);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip back
	}
}
microVUf(void) mVU_ESADD() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		getReg6(xmmFs, _Fs_);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip xmmPQ to get Valid P instance
		mVU_sumXYZ<vuIndex>();
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip back
	}
}
#define esinHelper(addr) {  \
	SSE_MULSS_XMM_to_XMM(xmmT1, xmmFt);  \
	SSE_MOVAPS_XMM_to_XMM(xmmFs, xmmT1);  \
	SSE_MULSS_M32_to_XMM(xmmFs, (uptr)addr);  \
	SSE_ADDSS_XMM_to_XMM(xmmPQ, xmmFs);  \
}
microVUf(void) mVU_ESIN() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		getReg5(xmmFs, _Fs_, _Fsf_);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip xmmPQ to get Valid P instance
		SSE_MOVSS_XMM_to_XMM(xmmPQ, xmmFs);
		//SSE_MULSS_M32_to_XMM(xmmPQ, (uptr)mVU_one); // Multiplying by 1 is redundant?
		SSE_MOVAPS_XMM_to_XMM(xmmFt, xmmFs);
		SSE_MULSS_XMM_to_XMM(xmmFs, xmmFt);
		SSE_MOVAPS_XMM_to_XMM(xmmT1, xmmFs);
		SSE_MULSS_XMM_to_XMM(xmmFs, xmmFt);
		SSE_MOVAPS_XMM_to_XMM(xmmFt, xmmFs);
		SSE_MULSS_M32_to_XMM(xmmFs, (uptr)mVU_S2);
		SSE_ADDSS_XMM_to_XMM(xmmPQ, xmmFs);

		esinHelper(mVU_S3);
		esinHelper(mVU_S4);
		
		SSE_MULSS_XMM_to_XMM(xmmT1, xmmFt);
		SSE_MULSS_M32_to_XMM(xmmT1, (uptr)mVU_S5);
		SSE_ADDSS_XMM_to_XMM(xmmPQ, xmmT1);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip back
	}
} 
microVUf(void) mVU_ESQRT() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		getReg5(xmmFs, _Fs_, _Fsf_);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip xmmPQ to get Valid P instance
		SSE_SQRTSS_XMM_to_XMM(xmmPQ, xmmFs);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip back
	}
}
microVUf(void) mVU_ESUM() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		getReg6(xmmFs, _Fs_);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip xmmPQ to get Valid P instance
		SSE2_PSHUFD_XMM_to_XMM(xmmFt, xmmFs, 0x1b);
		SSE_ADDPS_XMM_to_XMM(xmmFs, xmmFt);
		SSE2_PSHUFD_XMM_to_XMM(xmmFt, xmmFs, 0x01);
		SSE_ADDSS_XMM_to_XMM(xmmFs, xmmFt);
		SSE_MOVSS_XMM_to_XMM(xmmPQ, xmmFs);
		SSE2_PSHUFD_XMM_to_XMM(xmmPQ, xmmPQ, writeP ? 0x27 : 0xC6); // Flip back
	}
}

microVUf(void) mVU_FCAND() {}
microVUf(void) mVU_FCEQ() {}
microVUf(void) mVU_FCOR() {}
microVUf(void) mVU_FCSET() {}
microVUf(void) mVU_FCGET() {}

microVUf(void) mVU_FMAND() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		mVUallocMFLAGa<vuIndex>(gprT1, fvmInstance);
		mVUallocVIa<vuIndex>(gprT2, _Fs_);
		AND16RtoR(gprT1, gprT2);
		mVUallocVIb<vuIndex>(gprT1, _Ft_);
	}
}
microVUf(void) mVU_FMEQ() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		mVUallocMFLAGa<vuIndex>(gprT1, fvmInstance);
		mVUallocVIa<vuIndex>(gprT2, _Fs_);
		CMP16RtoR(gprT1, gprT2);
		SETE8R(gprT1);
		AND16ItoR(gprT1, 0x1);
		mVUallocVIb<vuIndex>(gprT1, _Ft_);
	}
}
microVUf(void) mVU_FMOR() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		mVUallocMFLAGa<vuIndex>(gprT1, fvmInstance);
		mVUallocVIa<vuIndex>(gprT2, _Fs_);
		OR16RtoR(gprT1, gprT2);
		mVUallocVIb<vuIndex>(gprT1, _Ft_);
	}
}

microVUf(void) mVU_FSAND() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		mVUallocSFLAGa<vuIndex>(gprT1, fvsInstance);
		AND16ItoR(gprT1, _Imm12_);
		mVUallocVIb<vuIndex>(gprT1, _Ft_);
	}
}
microVUf(void) mVU_FSEQ() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		mVUallocSFLAGa<vuIndex>(gprT1, fvsInstance);
		CMP16ItoR(gprT1, _Imm12_);
		SETE8R(gprT1);
		AND16ItoR(gprT1, 0x1);
		mVUallocVIb<vuIndex>(gprT1, _Ft_);
	}
}
microVUf(void) mVU_FSOR() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		mVUallocSFLAGa<vuIndex>(gprT1, fvsInstance);
		OR16ItoR(gprT1, _Imm12_);
		mVUallocVIb<vuIndex>(gprT1, _Ft_);
	}
}
microVUf(void) mVU_FSSET() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		int flagReg;
		getFlagReg(flagReg, fsInstance);
		MOV16ItoR(gprT1, (_Imm12_ & 0xfc0));
	}
}

microVUf(void) mVU_IADD() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		mVUallocVIa<vuIndex>(gprT1, _Fs_);
		if (_Ft_ != _Fs_) {
			mVUallocVIa<vuIndex>(gprT2, _Ft_);
			ADD16RtoR(gprT1, gprT2);
		}
		else ADD16RtoR(gprT1, gprT1);
		mVUallocVIb<vuIndex>(gprT1, _Fd_);
	}
}
microVUf(void) mVU_IADDI() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		mVUallocVIa<vuIndex>(gprT1, _Fs_);
		ADD16ItoR(gprT1, _Imm5_);
		mVUallocVIb<vuIndex>(gprT1, _Ft_);
	}
}
microVUf(void) mVU_IADDIU() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		mVUallocVIa<vuIndex>(gprT1, _Fs_);
		ADD16ItoR(gprT1, _Imm12_);
		mVUallocVIb<vuIndex>(gprT1, _Ft_);
	}
}
microVUf(void) mVU_IAND() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		mVUallocVIa<vuIndex>(gprT1, _Fs_);
		if (_Ft_ != _Fs_) {
			mVUallocVIa<vuIndex>(gprT2, _Ft_);
			AND32RtoR(gprT1, gprT2);
		}
		mVUallocVIb<vuIndex>(gprT1, _Fd_);
	}
}
microVUf(void) mVU_IOR() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		mVUallocVIa<vuIndex>(gprT1, _Fs_);
		if (_Ft_ != _Fs_) {
			mVUallocVIa<vuIndex>(gprT2, _Ft_);
			OR32RtoR(gprT1, gprT2);
		}
		mVUallocVIb<vuIndex>(gprT1, _Fd_);
	}
}
microVUf(void) mVU_ISUB() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		if (_Ft_ != _Fs_) {
			mVUallocVIa<vuIndex>(gprT1, _Fs_);
			mVUallocVIa<vuIndex>(gprT2, _Ft_);
			SUB16RtoR(gprT1, gprT2);
		}
		else if (!isMMX(_Fd_)) { 
			XOR32RtoR(gprT1, gprT1);
			mVUallocVIb<vuIndex>(gprT1, _Fd_);
		}
		else { PXORRtoR(mmVI(_Fd_), mmVI(_Fd_)); }
	}
}
microVUf(void) mVU_ISUBIU() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		mVUallocVIa<vuIndex>(gprT1, _Fs_);
		SUB16ItoR(gprT1, _Imm12_);
		mVUallocVIb<vuIndex>(gprT1, _Ft_);
	}
}

microVUf(void) mVU_B() {}
microVUf(void) mVU_BAL() {}
microVUf(void) mVU_IBEQ() {}
microVUf(void) mVU_IBGEZ() {}
microVUf(void) mVU_IBGTZ() {}
microVUf(void) mVU_IBLTZ() {}
microVUf(void) mVU_IBLEZ() {}
microVUf(void) mVU_IBNE() {}
microVUf(void) mVU_JR() {}
microVUf(void) mVU_JALR() {}

microVUf(void) mVU_MOVE() {
	microVU* mVU = mVUx;
	if (recPass == 0) { /*If (!_Ft_ || (_Ft_ == _Fs_)) nop();*/ }
	else { 
		mVUloadReg<vuIndex>(xmmT1, (uptr)&mVU->regs->VF[_Fs_].UL[0], _X_Y_Z_W);
		mVUsaveReg<vuIndex>(xmmT1, (uptr)&mVU->regs->VF[_Ft_].UL[0], _X_Y_Z_W);
	}
}
microVUf(void) mVU_MFIR() {
	microVU* mVU = mVUx;
	if (recPass == 0) { /*If (!_Ft_) nop();*/ }
	else { 
		mVUallocVIa<vuIndex>(gprT1, _Fs_);
		MOVSX32R16toR(gprT1, gprT1);
		SSE2_MOVD_R_to_XMM(xmmT1, gprT1);
		if (!_XYZW_SS) { mVUunpack_xyzw<vuIndex>(xmmT1, xmmT1, 0); }
		mVUsaveReg<vuIndex>(xmmT1, (uptr)&mVU->regs->VF[_Ft_].UL[0], _X_Y_Z_W);
	}
}
microVUf(void) mVU_MFP() {
	microVU* mVU = mVUx;
	if (recPass == 0) { /*If (!_Ft_) nop();*/ }
	else { 
		getPreg(xmmFt);
		mVUsaveReg<vuIndex>(xmmFt, (uptr)&mVU->regs->VF[_Ft_].UL[0], _X_Y_Z_W);
	}
}
microVUf(void) mVU_MTIR() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		MOVZX32M16toR(gprT1, (uptr)&mVU->regs->VF[_Fs_].UL[_Fsf_]);
		mVUallocVIb<vuIndex>(gprT1, _Ft_);
	}
}
microVUf(void) mVU_MR32() {
	microVU* mVU = mVUx;
	if (recPass == 0) { /*If (!_Ft_) nop();*/ }
	else { 
		mVUloadReg<vuIndex>(xmmT1, (uptr)&mVU->regs->VF[_Fs_].UL[0], (_X_Y_Z_W == 8) ? 4 : 15);
		if (_X_Y_Z_W != 8) { SSE2_PSHUFD_XMM_to_XMM(xmmT1, xmmT1, 0x39); }
		mVUsaveReg<vuIndex>(xmmT1, (uptr)&mVU->regs->VF[_Ft_].UL[0], _X_Y_Z_W);
	}
}

microVUf(void) mVU_ILW() {
	microVU* mVU = mVUx;
	if (recPass == 0) { /*If (!_Ft_) nop();*/ }
	else { 
		if (!_Fs_) {
			MOVZX32M16toR( gprT1, (uptr)mVU->regs->Mem + getVUmem(_Imm11_) + offsetSS );
			mVUallocVIb<vuIndex>(gprT1, _Ft_);
		}
		else {
			mVUallocVIa<vuIndex>(gprT1, _Fs_);
			ADD32ItoR(gprT1, _Imm11_);
			mVUaddrFix<vuIndex>(gprT1);
			MOV32RmSOffsettoR(gprT1, gprT1, (uptr)mVU->regs->Mem + offsetSS, 0); // ToDo: check if this works.
			if (isMMX(_Ft_)) AND32ItoR(gprT1, 0xffff);
			mVUallocVIb<vuIndex>(gprT1, _Ft_);
		}
	}
}
microVUf(void) mVU_ILWR() {
	microVU* mVU = mVUx;
	if (recPass == 0) { /*If (!_Ft_) nop();*/ }
	else { 
		if (!_Fs_) {
			MOVZX32M16toR( gprT1, (uptr)mVU->regs->Mem + offsetSS );
			mVUallocVIb<vuIndex>(gprT1, _Ft_);
		}
		else {
			mVUallocVIa<vuIndex>(gprT1, _Fs_);
			mVUaddrFix<vuIndex>(gprT1);
			MOV32RmSOffsettoR(gprT1, gprT1, (uptr)mVU->regs->Mem + offsetSS, 0); // ToDo: check if this works.
			if (isMMX(_Ft_)) AND32ItoR(gprT1, 0xffff);
			mVUallocVIb<vuIndex>(gprT1, _Ft_);
		}
	}
}
microVUf(void) mVU_ISW() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		if (!_Fs_) {
			int imm = getVUmem(_Imm11_);
			mVUallocVIa<vuIndex>(gprT1, _Ft_);
			if (_X) MOV32RtoM((uptr)mVU->regs->Mem + imm,		gprT1);
			if (_Y) MOV32RtoM((uptr)mVU->regs->Mem + imm + 4,	gprT1);
			if (_Z) MOV32RtoM((uptr)mVU->regs->Mem + imm + 8,	gprT1);
			if (_W) MOV32RtoM((uptr)mVU->regs->Mem + imm + 12,	gprT1);
		}
		else {
			mVUallocVIa<vuIndex>(gprT1, _Fs_);
			mVUallocVIa<vuIndex>(gprT2, _Ft_);
			ADD32ItoR(gprT1, _Imm11_);
			mVUaddrFix<vuIndex>(gprT1);
			if (_X) MOV32RtoRmOffset(gprT1, gprT2, (uptr)mVU->regs->Mem);
			if (_Y) MOV32RtoRmOffset(gprT1, gprT2, (uptr)mVU->regs->Mem+4);
			if (_Z) MOV32RtoRmOffset(gprT1, gprT2, (uptr)mVU->regs->Mem+8);
			if (_W) MOV32RtoRmOffset(gprT1, gprT2, (uptr)mVU->regs->Mem+12);
		}
	}
}
microVUf(void) mVU_ISWR() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		if (!_Fs_) {
			mVUallocVIa<vuIndex>(gprT1, _Ft_);
			if (_X) MOV32RtoM((uptr)mVU->regs->Mem,	   gprT1);
			if (_Y) MOV32RtoM((uptr)mVU->regs->Mem+4,  gprT1);
			if (_Z) MOV32RtoM((uptr)mVU->regs->Mem+8,  gprT1);
			if (_W) MOV32RtoM((uptr)mVU->regs->Mem+12, gprT1);
		}
		else {
			mVUallocVIa<vuIndex>(gprT1, _Fs_);
			mVUallocVIa<vuIndex>(gprT2, _Ft_);
			mVUaddrFix<vuIndex>(gprT1);
			if (_X) MOV32RtoRmOffset(gprT1, gprT2, (uptr)mVU->regs->Mem);
			if (_Y) MOV32RtoRmOffset(gprT1, gprT2, (uptr)mVU->regs->Mem+4);
			if (_Z) MOV32RtoRmOffset(gprT1, gprT2, (uptr)mVU->regs->Mem+8);
			if (_W) MOV32RtoRmOffset(gprT1, gprT2, (uptr)mVU->regs->Mem+12);
		}
	}
}

microVUf(void) mVU_LQ() {
	microVU* mVU = mVUx;
	if (recPass == 0) { /*If (!_Ft_) nop();*/ }
	else { 
		if (!_Fs_) {
			mVUloadReg<vuIndex>(xmmFt, (uptr)mVU->regs->Mem + getVUmem(_Imm11_), _X_Y_Z_W);
			mVUsaveReg<vuIndex>(xmmFt, (uptr)&mVU->regs->VF[_Ft_].UL[0], _X_Y_Z_W);
		}
		else {
			mVUallocVIa<vuIndex>(gprT1, _Fs_);
			ADD32ItoR(gprT1, _Imm11_);
			mVUaddrFix<vuIndex>(gprT1);
			mVUloadReg2<vuIndex>(xmmFt, gprT1, (uptr)mVU->regs->Mem, _X_Y_Z_W);
			mVUsaveReg<vuIndex>(xmmFt, (uptr)&mVU->regs->VF[_Ft_].UL[0], _X_Y_Z_W);
		}
	}
}
microVUf(void) mVU_LQD() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		if (!_Fs_ && _Ft_) {
			mVUloadReg<vuIndex>(xmmFt, (uptr)mVU->regs->Mem, _X_Y_Z_W);
			mVUsaveReg<vuIndex>(xmmFt, (uptr)&mVU->regs->VF[_Ft_].UL[0], _X_Y_Z_W);
		}
		else {
			mVUallocVIa<vuIndex>(gprT1, _Fs_);
			SUB16ItoR(gprT1, 1);
			mVUallocVIb<vuIndex>(gprT1, _Fs_); // ToDo: Backup to memory check.
			if (_Ft_) {
				mVUaddrFix<vuIndex>(gprT1);
				mVUloadReg2<vuIndex>(xmmFt, gprT1, (uptr)mVU->regs->Mem, _X_Y_Z_W);
				mVUsaveReg<vuIndex>(xmmFt, (uptr)&mVU->regs->VF[_Ft_].UL[0], _X_Y_Z_W);
			}
		}
	}
}
microVUf(void) mVU_LQI() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		if (!_Fs_ && _Ft_) {
			mVUloadReg<vuIndex>(xmmFt, (uptr)mVU->regs->Mem, _X_Y_Z_W);
			mVUsaveReg<vuIndex>(xmmFt, (uptr)&mVU->regs->VF[_Ft_].UL[0], _X_Y_Z_W);
		}
		else {
			mVUallocVIa<vuIndex>((_Ft_) ? gprT1 : gprT2, _Fs_);
			if (_Ft_) {
				MOV32RtoR(gprT2, gprT1);
				mVUaddrFix<vuIndex>(gprT1);
				mVUloadReg2<vuIndex>(xmmFt, gprT1, (uptr)mVU->regs->Mem, _X_Y_Z_W);
				mVUsaveReg<vuIndex>(xmmFt, (uptr)&mVU->regs->VF[_Ft_].UL[0], _X_Y_Z_W);
			}
			ADD16ItoR(gprT2, 1);
			mVUallocVIb<vuIndex>(gprT2, _Fs_); // ToDo: Backup to memory check.
		}
	}
}
microVUf(void) mVU_SQ() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		if (!_Ft_) {
			getReg7(xmmFs, _Fs_);
			mVUsaveReg<vuIndex>(xmmFs, (uptr)mVU->regs->Mem + getVUmem(_Imm11_), _X_Y_Z_W);
		}
		else {
			mVUallocVIa<vuIndex>(gprT1, _Ft_);
			ADD32ItoR(gprT1, _Imm11_);
			mVUaddrFix<vuIndex>(gprT1);
			getReg7(xmmFs, _Fs_);
			mVUsaveReg2<vuIndex>(xmmFs, gprT1, (uptr)mVU->regs->Mem, _X_Y_Z_W);
		}
	}
}
microVUf(void) mVU_SQD() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		if (!_Ft_) {
			getReg7(xmmFs, _Fs_);
			mVUsaveReg<vuIndex>(xmmFs, (uptr)mVU->regs->Mem, _X_Y_Z_W);
		}
		else {
			mVUallocVIa<vuIndex>(gprT1, _Ft_);
			SUB16ItoR(gprT1, 1);
			mVUallocVIb<vuIndex>(gprT1, _Ft_); // ToDo: Backup to memory check.
			mVUaddrFix<vuIndex>(gprT1);
			getReg7(xmmFs, _Fs_);
			mVUsaveReg2<vuIndex>(xmmFs, gprT1, (uptr)mVU->regs->Mem, _X_Y_Z_W);
		}
	}
}
microVUf(void) mVU_SQI() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		if (!_Ft_) {
			getReg7(xmmFs, _Fs_);
			mVUsaveReg<vuIndex>(xmmFs, (uptr)mVU->regs->Mem, _X_Y_Z_W);
		}
		else {
			mVUallocVIa<vuIndex>(gprT1, _Ft_);
			MOV32RtoR(gprT2, gprT1);
			mVUaddrFix<vuIndex>(gprT1);
			getReg7(xmmFs, _Fs_);
			mVUsaveReg2<vuIndex>(xmmFs, gprT1, (uptr)mVU->regs->Mem, _X_Y_Z_W);
			ADD16ItoR(gprT2, 1);
			mVUallocVIb<vuIndex>(gprT2, _Ft_); // ToDo: Backup to memory check.
		}
	}
}

microVUf(void) mVU_RINIT() {}
microVUf(void) mVU_RGET() {}
microVUf(void) mVU_RNEXT() {}
microVUf(void) mVU_RXOR() {}

microVUf(void) mVU_WAITP() {}
microVUf(void) mVU_WAITQ() {}

microVUf(void) mVU_XGKICK() {}
microVUf(void) mVU_XTOP() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		MOVZX32M16toR( gprT1, (uptr)&mVU->regs->vifRegs->top);
		mVUallocVIb<vuIndex>(gprT1, _Ft_);
	}
}
microVUf(void) mVU_XITOP() {
	microVU* mVU = mVUx;
	if (recPass == 0) {}
	else { 
		MOVZX32M16toR( gprT1, (uptr)&mVU->regs->vifRegs->itop );
		mVUallocVIb<vuIndex>(gprT1, _Ft_);
	}
}
#endif //PCSX2_MICROVU
